{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpGlobals.pas. The Original Code is available through  }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit defines global variables, constants, types, etc. for use with      }
{ DSpatial. Most importantly, this unit contains the TDSpatial class, and its  }
{ DSpatial instance, which controls the global behaviour of any application    }
{ based on DSpatial. This unit should be included in virtually every other unit}
{ that is used in DSpatial and it should be the first listed DSpatial unit in  }
{ the uses clause of the project file (*.DPR).                                 }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 August 14                                               }
{                                                                              }
{ Dependencies  : DSpUtils (TFileMapping)                                      }
{                 Messages (WM_APP)                                            }
{                 SysUtils (Format)                                            }
{                                                                              }
{ Version       : 0.3  03Aug14 Conversion of TDSpUnit to a class               }
{                              Addition of TDSpRasterRenderMap                 }
{                 0.2  03Jul18 Addition of TDSpatial                           }
{                              Changes/Additions to TDSpExtent                 }
{                 0.1  03Apr26 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpGlobals;
interface
uses Windows, Messages, DSpUtils;
const
  // Global constants
  DSP_MAXDATASIZE       = $60000000;  // 1.5GB, maximum allocation of memory
                                      // for data sets

  // Message IDs for message passing and thread management
  DSPMSG_DSPBASE        = WM_APP + 400;
  DSPMSG_TERMINATE      = DSPMSG_DSPBASE + 1;

  DSPMSG_PFABASE        = DSPMSG_DSPBASE + 200;
  DSPMSG_DATASOURCEINFO = DSPMSG_PFABASE + 1;
  DSPMSG_DATASETINFO    = DSPMSG_PFABASE + 10;
  DSPMSG_RASTERINFO     = DSPMSG_PFABASE + 12;
  DSPMSG_BANDINFO       = DSPMSG_PFABASE + 13;
  DSPMSG_BANDREAD       = DSPMSG_PFABASE + 14;
  DSPMSG_BANDREADNATIVE = DSPMSG_PFABASE + 15;
  DSPMSG_BANDREADBLOCK  = DSPMSG_PFABASE + 16;
  DSPMSG_BANDWRITE      = DSPMSG_PFABASE + 21;
  DSPMSG_VECTORINFO     = DSPMSG_PFABASE + 41;

type
  // Data types. For spatial data these identifiers must be used. For any other
  // persistent data these types are highly recommended.
  TDSpUInt8    = byte;
  TDSpUInt16   = word;
  TDSpUInt32   = cardinal;
  TDSpInt8     = smallint;
  TDSpInt16    = shortint;
  TDSpInt32    = integer;
  TDSpInt64    = Int64;
  TDSpFloat32  = single;
  TDSpFloat64  = double;

  TDSpDataType = (dspUInt8, dspUInt16, dspUInt32, dspInt8, dspInt16,
                  dspInt32, dspInt64, dspFloat32, dspFloat64);

  // Object states or results. These can be used by many different objects.
  TDSpState = (dspOk, dspError, dspIgnore, dspWarning, dspFail, dspPending,
               dspIdle, dspBusy, dspReading, dspWriting, dspProcessing,
               dspRendering, dspDeleting);
  TDSpStates = set of TDSpState;
  
  // PFA types
  TDSpPFACap = (dspReadRaster, dspWriteRaster, dspReadVector, dspWriteVector,
                dspReadAttribute, dspWriteAttribute);
  TDSpPFACaps = set of TDSpPFACap;

  TDSpPFAFormat = (dspDontKnow, dspRasterData, dspVectorData);

  TDSpRasterBandType = (dspRasterBandSurface, dspRasterBandOrdinal);

  TDSpDataReadType = (dspReadNative, dspReadBand, dspReadBandArea);

  TDSpRasterBandFormat = (dspRasterFlat, dspRasterInterleaveLine, dspRasterInterleavePixel,
                          dspRasterTiled);

  TDSpPFAOperation = (dspNull, dspInitReader, dspDescribeDataSets, dspReadInfo,
                      dspReadRasterInfo, dspReadRasterBandInfo,
                      dspReadRasterBandFull, dspReadRasterBandRows, dspReadRasterBandRegion,
                      dspreadRasterBandTile);

  // Other stuff
  TParameterEvent = procedure (Sender: TObject; p: double) of object;

  TArraySingle  = array [0..(MaxInt div SizeOf(single) - 1)] of single;
  PArraySingle  = ^TArraySingle;

  TDSpDescribe  = (dspFriendly, dspWKT);
  TDSpDirection = (dspHorz, dspVert);

  // TDSpatial
  // Never instantiate this class. Use the global DSpatial variable for access
  TDSpatial = class
  private
//    FMap       : TFileMapping;
//    FParams    : TFileView;
  protected
    constructor Create;
  public
    destructor  Destroy; override;

    function    NewDataSetID: TDSpUInt32;
  end;

  // Extent management and manipulation
  // This class will not function properly with lat/long coordinate systems
  // that span the date line and/or poles.
  TDSpExtent = class
  protected
    FNorth, FEast, FSouth, FWest : TDSpFloat64;
  private
    function GetArea: TDSpFloat64;
  public
    constructor Create(north, east, south, west: TDSpFloat64);

    function    Clone: TDSpExtent;
    procedure   CopyFrom(ex: TDSpExtent);
    procedure   UnionWith(ex: TDSpExtent);
    function    Intersects(ex: TDSpExtent): boolean; overload;
    function    Intersects(n, e, s, w: TDSpFloat64): boolean; overload;
    property    Area: TDSpFloat64 read GetArea;
    property    West: TDSpFloat64 read FWest write FWest;
    property    East: TDSpFloat64 read FEast write FEast;
    property    North: TDSpFloat64 read FNorth write FNorth;
    property    South: TDSpFloat64 read FSouth write FSouth;
  end;

  // Units of measure
  TDSpUnit  = class
  protected
    FName   : string;
    FAbbrev : string;
    FConvert: double;
  public
    constructor Create(name: string; convert: double);

    function    Describe(descr: TDSpDescribe): string;
    function    IsCompatible(u: TDSpUnit): boolean;
    property    Name: string read FName write FName;
    property    Convert: double read FConvert;
    property    Short: string read FAbbrev write FAbbrev;
  end;

  // Raster render map
  TDSpWordMap = ^word;

  TDSpRasterRenderMap = class
  private
    FSize: integer;
  protected
    FMap: TDSpWordMap;
    FRows, FColumns: integer;

    function    GetRow(r: integer): TDSpWordMap;
  public
    constructor Create(r, c: integer);
    destructor  Destroy; override;

    procedure   Resize(r, c: integer);
    procedure   Pack;

    property    Row[i: integer]: TDSpWordMap read GetRow;
  end;

  function DataSize(t: TDSpDataType): TDSpInt32;

var
  DSpatial: TDSpatial;

implementation
// Load the necesary resources
(*  {$R DSpatial.res}        *)
uses SysUtils;

////////////////////////////////////////////////////////////////////////////////
// TDSpatial
constructor TDSpatial.Create;
begin
end;

destructor TDSpatial.Destroy;
begin
end;

function TDSpatial.NewDataSetID: TDSpUInt32;
begin
  Result := 0; // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpExtent
constructor TDSpExtent.Create(north, east, south, west: TDSpFloat64);
begin
  FEast  := east;
  FWest  := west;
  FNorth := north;
  FSouth := south;
end;

function TDSpExtent.Clone: TDSpExtent;
begin
  Result := TDSpExtent.Create(FNorth, FEast, FSouth, FWest);
end;

procedure TDSpExtent.CopyFrom(ex: TDSpExtent);
begin
  FEast := ex.FEast;
  FWest := ex.FWest;
  FNorth := ex.FNorth;
  FSouth := ex.FSouth;
end;

procedure TDSpExtent.UnionWith(ex: TDSpExtent);
begin
  if ex.FEast > FEast then FEast := ex.FEast;
  if ex.FWest < FWest then FWest := ex.FWest;
  if ex.FNorth > FNorth then FNorth := ex.FNorth;
  if ex.FSouth < FSouth then FSouth := ex.FSouth;
end;

function TDSpExtent.Intersects(ex: TDSpExtent): boolean;
begin
  Result := false; // always the pessimist...
  if FWest < ex.West then begin
    if FEast < ex.West then Exit;
  end else if ex.East < FWest then Exit;
  if FSouth < ex.South then begin
    if FNorth < ex.South then Exit;
  end else if ex.North < FSouth then Exit;
  Result := true;
end;

function TDSpExtent.Intersects(n, e, s, w: TDSpFloat64): boolean;
begin
  Result := false; // always the pessimist...
  if FWest < w then begin
    if FEast < w then Exit;
  end else if e < FWest then Exit;
  if FSouth < s then begin
    if FNorth < s then Exit;
  end else if n < FSouth then Exit;
  Result := true;
end;

function TDSpExtent.GetArea: TDSpFloat64;
begin
  Result := (FNorth - FSouth) * (FEast - FWest);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpUnit
constructor TDSpUnit.Create(name: string; convert: double);
begin
  inherited Create;
  FName := name;
  FConvert := convert;
end;

function TDSpunit.Describe(descr: TDSpDescribe): string;
begin
  if descr = dspWKT then Result := Format('UNIT["%s",%15.13f]', [FName, FConvert])
  else Result := '  Unit: ' + FName + Format(' (%15.13f ', [FConvert]);
end;

function TDSpunit.IsCompatible(u: TDSpUnit): boolean;
begin
  Result := FloatsSame(FConvert, u.Convert);
end;

////////////////////////////////////////////////////////////////////////////////
//
constructor TDSpRasterRenderMap.Create(r, c: integer);
begin
  inherited Create;
  FRows := r;
  FColumns := c;
  FSize := r * c;
  FMap := AllocMem(FSize * 2);
end;

destructor TDSpRasterRenderMap.Destroy;
begin
  FreeMem(FMap);
  inherited;
end;

procedure TDSpRasterRenderMap.Resize(r, c: integer);
begin
  FRows := r;
  FColumns := c;
  if FRows * FColumns > FSize then begin
    FSize := FRows * FColumns;
    ReallocMem(FMap, FSize * 2);
  end;
  FillChar(FMap^, FSize * 2, 0);
end;

procedure TDSpRasterRenderMap.Pack;
begin
  if FRows * FColumns < FSize then begin
    FSize := FRows * FColumns;
    ReallocMem(FMap, FSize * 2);
  end;
end;

function TDSpRasterRenderMap.GetRow(r: integer): TDSpWordMap;
begin
  Result := pointer(integer(FMap) + r * FColumns * 2);
end;

////////////////////////////////////////////////////////////////////////////////
// Global functions
function DataSize(t: TDSpDataType): TDSpInt32;
begin
  case t of
    dspUInt8, dspInt8               : Result := 1;
    dspUInt16, dspInt16             : Result := 2;
    dspUInt32, dspInt32, dspFloat32 : Result := 4;
    dspInt64, dspFloat64            : Result := 8;
  else Result := 0;            // satisfy the compiler
  end;
end;

initialization
  DSpatial := TDSpatial.Create;
finalization
  DSpatial.Free;
end.
