// Delaunay triangulation, Voronoi diagram

unit dtmDvUnit;

interface

uses
  Windows, Messages, SysUtils, Classes,
  Graphics, Controls, Forms, Dialogs;

type

  TDPoint = class;
  TDEdge = class;
  TDTriangle = class;
  TDelaunay = class;

  TEnumPointEvent = procedure(Sender: TObject; Point: TDPoint) of object;
  TEnumEdgeEvent = procedure(Sender: TObject; Edge: TDEdge) of object;
  TEnumTriEvent = procedure(Sender: TObject; Triangle: TDTriangle) of object;

  TDPoint = class
  private
    FOwner: TDelaunay;
    FX, FY: Integer;    // coordinate X,Y
    FE: TDEdge;         // an edge which start from this Point
    FOriginal: TDPoint;
    procedure SetOwner(Owner: TDelaunay);
    function GetE: TDEdge;
    function GetIndex: Integer;
    function GetIsOriginal: Boolean;
    procedure Clear;
    function GetFirstE: TDEdge;
    function GetLastE: TDEdge;
    function GetCount: Integer;
    function GetNextOrig: TDPoint;
    function GetPriorOrig: TDPoint;
  public
    constructor Create(X, Y: Integer);
    property E: TDEdge read GetE;
    property X: Integer read FX;
    property Y: Integer read FY;
    function Distance(PX, PY: Double): Double;
    property Index: Integer read GetIndex;
    property Owner: TDelaunay read FOwner write SetOwner;
    destructor Destroy; override;
    property Original: TDPoint read FOriginal;
    property IsOriginal: Boolean read GetIsOriginal;
    function IsIdent(P: TDPoint): Boolean;
    property FirstE: TDEdge read GetFirstE;
    property LastE: TDEdge read GetLastE;
    procedure EnumEdges(Event: TEnumEdgeEvent);
    property Count: Integer read GetCount;
    procedure Move(NewX, NewY: Integer);
    property NextOrig: TDPoint read GetNextOrig;
    property PriorOrig: TDPoint read GetPriorOrig;
  end;

  TDEdge = class
  private
    FLP, FRP: TDPoint; // start and end point of the edge
    A, B, C: Double;  // line equation parameters. aX+bY+c=0
    FT: TDTriangle;   // triangle containing this edge
    FNextE: TDEdge;   // next edge in the triangle in countclickwise
    FNextH: TDEdge;   // convex hull link
    FInvE: TDEdge;    // inverse edge (p2->p1)
    constructor Create(LP, RP: TDPoint);
    procedure Update(LP, RP: TDPoint);
    procedure ChangeP(OldP, NewP: TDPoint);
    function  CreateInv: TDEdge;
    procedure SetInvE(e: TDEdge);
    procedure Setabc;   // set parameters of a,b,c
    procedure SetPointRefer;
    function  GetNextPE: TDEdge;
  public
    property LeftP: TDPoint read FLP;
    property RightP: TDPoint read FRP;
    property InvE: TDEdge read FInvE;
    property NextE: TDEdge read FNextE;
    property NextH: TDEdge read FNextH;
    property T: TDTriangle read FT;
    property ParamA: Double read A;
    property ParamB: Double read B;
    property ParamC: Double read C;
    function IsPrior: Boolean;
    function OnSide(P: TDPoint): Integer;
    function InSection(P: TDPoint): Integer;
    property NextPE: TDEdge read GetNextPE;
    procedure GetVoronoi(var X1, Y1, X2, Y2: Double);
  end;

  TDTriangle = class
  private
    FE: TDEdge;          // an edge of this triangle
    FCX: Double;         // center of circle: X
    FCY: Double;         // center of circle: Y
    FCR: Double;         // radius of circle
    Index: Integer;
    FPriorT, FNextT: TDTriangle;
    constructor Create(E1, E2, E3: TDEdge);
    procedure Update(E1, E2, E3: TDEdge);
    procedure FindCircle;
  public
    property E: TDEdge read FE;
    property CX: Double read FCX;
    property CY: Double read FCY;
    property CR: Double read FCR;
    property PriorT: TDTriangle read FPriorT;
    property NextT: TDTriangle read FNextT;
    function InCircle(P: TDPoint): Boolean;
    function IsIn(P: TDPoint): Boolean;
  end;

  TDelaunay = class(TComponent)
  private
    Points: TList;        // points set
    FHullStart: TDEdge;   // entring edge of convex hull
    FFirstT: TDTriangle;
    FOnEnumPoint: TEnumPointEvent;
    FOnEnumEdge: TEnumEdgeEvent;
    FOnEnumHEdge: TEnumEdgeEvent;
    FOnEnumTrinagle: TEnumTriEvent;
    FOnInsert: TEnumPointEvent;
    FOnRemove: TEnumPointEvent;
    FBreakEnum: Boolean;
    procedure UpdateTri(T: TDTriangle; P: TDPoint);
    procedure UpdateHull(P: TDPoint; Ind: Integer);
    procedure SwapTri(e11: TDEdge);
    procedure DeleteTri(T: TDTriangle);
    procedure AddTri(T: TDTriangle);
    function FindTri(P: TDPoint): TDTriangle;
    function GetPoint(Index: Integer): TDPoint;
    function GetPointsCount: Integer;
    procedure Delete(P: TDPoint);
    procedure DeleteOnHull(StartE: TDEdge);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property Count: Integer read GetPointsCount;
    property Point[Index: Integer]: TDPoint read GetPoint; default;
    property FirstT: TDTriangle read FFirstT;
    property HullStart: TDEdge read FHullStart;
    function RightOrder(X, Y: Integer): Integer;
    function ByCoord(X, Y: Integer): TDPoint;
    function Nearest(X, Y: Integer): TDPoint;
    function Exists(X, Y: Integer): Boolean;
    function FindTriangle(X, Y: Integer): TDTriangle;

    procedure Clear(DelP: Boolean);
    function  Insert(px, py: Integer): TDPoint;
    procedure InsertPoint(P: TDPoint);
    procedure RemovePoint(P: TDPoint);
    procedure DeleteDupl;
    procedure GetPList(L: TList);

    procedure BreakEnum;
    procedure EnumPoints(Dupl: Boolean);
    procedure EnumEdges(BothSide: Boolean);
    procedure EnumTriangles;
    procedure EnumHEdges;
  published
    property OnEnumPoint: TEnumPointEvent read FOnEnumPoint write FOnEnumPoint;
    property OnEnumEdge: TEnumEdgeEvent read FOnEnumEdge write FOnEnumEdge;
    property OnEnumHEdge: TEnumEdgeEvent read FOnEnumHEdge write FOnEnumHEdge;
    property OnEnumTriangle: TEnumTriEvent read FOnEnumTrinagle write FOnEnumTrinagle;
    property OnInsert: TEnumPointEvent read FOnInsert write FOnInsert;
    property OnRemove: TEnumPointEvent read FOnRemove write FOnRemove;
  end;

procedure Register;

implementation

uses Consts; //Borland Strings

// TDPoint

constructor TDPoint.Create(X, Y: Integer);
begin
  FX := X;
  FY := Y;
  Clear;
end;

procedure TDPoint.Clear;
begin
  FE := nil;
  FOriginal := Self;
  FOwner := nil;
end;

function TDPoint.GetE: TDEdge;
begin
  Result := Original.FE;
end;

function TDPoint.GetIndex: Integer;
begin
  if FOwner = nil then
    Result := -1
  else
    with Owner, Points do
    begin
      Result := RightOrder(X, Y);
      while List[Result] <> Self do inc(Result);
    end;
end;

function TDPoint.GetCount: Integer;
var
  P: TDPoint;
  i: Integer;
begin
  Result := 1;
  if Owner <> nil then
    with Owner.Points do
      for i := Original.Index + 1 to Count - 1 do
      begin
	P := List[i];
	if P.Original = P then Break else inc(Result);
      end;
end;

function TDPoint.GetPriorOrig: TDPoint;
var
  i: Integer;
begin
  i := Original.Index;
  if i > 0 then
    Result := Owner.Points.List[i - 1]
  else
    Result := nil;
end;

function TDPoint.GetNextOrig: TDPoint;
var
  P: TDPoint;
  i: Integer;
begin
  Result := nil;
  if Owner <> nil then
    with Owner.Points do
      for i := Index + 1 to Count - 1 do
      begin
	P := List[i];
	if P.Original = P then
	begin
	  Result := P;
	  Break
	end;
      end;
end;

function TDPoint.GetIsOriginal: Boolean;
begin
  Result := Original = Self;
end;

function TDPoint.IsIdent(P: TDPoint): Boolean;
begin
  Result := (X = P.X) and (Y = P.Y);
end;

procedure TDPoint.SetOwner(Owner: TDelaunay);
begin
  if FOwner <> Owner then
  begin
    if FOwner <> nil then FOwner.RemovePoint(Self);
    if Owner <> nil then Owner.InsertPoint(Self);
  end;
end;

function TDPoint.Distance(px, py: Double): Double;
begin
  Result := sqrt(sqr(px - x) + sqr(py - y));
end;

function TDPoint.GetFirstE: TDEdge;
var SE: TDEdge;
begin
  Result := E;
  if Result = nil then Exit;
  SE := Result;
  while (Result.InvE <> nil) do
  begin
    Result := Result.InvE.NextE;
    if Result = SE then Break;
  end;
end;

function TDPoint.GetLastE: TDEdge;
var SE: TDEdge;
begin
  Result := E;
  if Result <> nil then
    if Result.T = nil then
    begin
      if (Result.RightP <> Original) and (Result.NextE <> nil) then
        Result := Result.NextE;
    end else begin
      SE := Result;
      repeat
        Result := Result.NextE.NextE;
        if Result.InvE = nil then Exit;
        Result := Result.InvE;
      until Result = SE;
      Result := Result.InvE.NextE;
    end;
end;

procedure TDPoint.EnumEdges;
var
  E, SE: TDEdge;
begin
  if not Assigned(Event) then Exit;
  E := FirstE;
  if E = nil then Exit;
  SE := LastE;
  Owner.FBreakEnum := False;
  repeat
    Event(Self, E);
    if (E = SE) or Owner.FBreakEnum then Break;
    E := E.NextPE;
  until False;
end;

procedure TDPoint.Move(NewX, NewY: Integer);
var Owner: TDelaunay;
begin
  if (NewX = FX) and (NewY = FY) then Exit;
  Owner := FOwner;
  SetOwner(nil);
  FX := NewX; FY := NewY;
  SetOwner(Owner);
end;

destructor TDPoint.Destroy;
begin
  SetOwner(nil);
end;

// TDEdge

constructor TDEdge.Create(LP, RP: TDPoint);
begin
  Update(LP, RP);
end;

procedure TDEdge.Update(LP, RP: TDPoint);
begin
  FLP := LP; FRP := RP;
  SetABC;
  SetPointRefer;
end;

procedure TDEdge.ChangeP(OldP, NewP: TDPoint);
begin
  if LeftP = OldP then FLP := NewP;
  if RightP = OldP then FRP := NewP;
  SetABC;
  SetPointRefer;
end;

function TDEdge.CreateInv: TDEdge;
begin
  Result := TDEdge.Create(RightP, LeftP);
  SetInvE(Result);
end;

procedure TDEdge.SetInvE(E: TDEdge);
begin
  FInvE := E;
  if E <> nil then E.FInvE := Self;
end;

function TDEdge.OnSide(P: TDPoint): Integer;
var s: Double;
begin
  s := A * P.X + B * P.Y + C;
  if s > 0 then Result := 1 else
    if s < 0 then Result := -1 else
      Result := 0;
end;

function TDEdge.InSection(P: TDPoint): Integer;
var s: Double;
begin
  if B = 0 then
    s := (P.Y - LeftP.Y) / A
  else
    s := (P.X - LeftP.X) / -B;
  if s > 1 then Result := 1 else
    if s < 0 then Result := -1 else
      Result := 0;
end;

procedure TDEdge.SetABC;
begin
   A := RightP.Y - LeftP.Y;
   B := LeftP.X - RightP.X;
   C := RightP.X * LeftP.Y - LeftP.X * RightP.Y;
end;

procedure TDEdge.SetPointRefer;
begin
  LeftP.FE := Self;
end;

function TDEdge.IsPrior: Boolean;
begin
  Result := (InvE = nil) or (InvE.T.Index > T.Index);
end;

function TDEdge.GetNextPE: TDEdge;
begin
  if T = nil then
    Result := NextE
  else begin
    Result := NextE.NextE;
    if Result.InvE <> nil then Result := Result.InvE;
  end;
end;

procedure TDEdge.GetVoronoi(var X1, Y1, X2, Y2: Double);
begin
  if T = nil then
  begin
    X1 := (LeftP.X + RightP.X) / 2 + A;
    X2 := X1 - (A + A);
    Y1 := (LeftP.Y + RightP.Y) / 2 + B;
    Y2 := Y1 - (B + B);
  end else
    with T do begin
      if (InvE = nil) then
      begin
        X1 := CX - A; Y1 := CY - B;
      end else
        with InvE.T do
        begin
          X1 := CX; Y1 := CY;
        end;
      X2 := CX; Y2 := CY;
    end;
end;

// TDTriangle

constructor TDTriangle.Create(e1, e2, e3: TDEdge);
begin
  Update(e1, e2, e3);
end;

procedure TDTriangle.Update(e1, e2, e3: TDEdge);
begin
  FE := e1;
  e1.FNextE := e2;
  e2.FNextE := e3;
  e3.FNextE := e1;
  e1.FT := Self;
  e2.FT := Self;
  e3.FT := Self;
  FindCircle;
end;

function TDTriangle.InCircle(P: TDPoint): Boolean;
begin
  Result := P.Distance(CX, CY) < CR;
end;

procedure TDTriangle.FindCircle;
var
  x1, y1, x2, y2, x3, y3, a, a1, a2: Double;
begin
  x1 := E.LeftP.X;
  y1 := E.LeftP.Y;
  x2 := E.RightP.X;
  y2 := E.RightP.Y;
  x3 := E.NextE.RightP.X;
  y3 := E.NextE.RightP.Y;
  a  := (y2 - y3) * (x2 - x1) - (y2 - y1) * (x2 - x3);
  a1 := (x1 + x2) * (x2 - x1) + (y2 - y1) * (y1 + y2);
  a2 := (x2 + x3) * (x2 - x3) + (y2 - y3) * (y2 + y3);
  FCX := (a1 * (y2 - y3) - a2 * (y2 - y1)) / a / 2;
  FCY := (a2 * (x2 - x1) - a1 * (x2 - x3)) / a / 2;
  FCR := E.LeftP.Distance(CX, CY);
end;

function TDTriangle.IsIn(P: TDPoint): Boolean;
begin
  Result := (E.OnSide(P) <> -1) and
            (E.NextE.OnSide(P) <> -1) and
            (E.NextE.NextE.OnSide(P) <> -1);
end;

// TDelaunay

constructor TDelaunay.Create(AOwner: TComponent);
begin
  inherited;
  Points := TList.Create;
end;

destructor TDelaunay.Destroy;
begin
  Clear(True);
  Points.Destroy;
  inherited;
end;

procedure TDelaunay.Clear;
var
  i: Integer;
  T: TDTriangle;
  E: TDEdge;
begin
  if FirstT = nil then
    while HullStart <> nil do
    begin
      E := HullStart;
      FHullStart := E.NextE;
      E.Destroy;
    end
  else
    while FirstT <> nil do
    begin
      T := FirstT;
      with T do
      begin
        E.NextE.NextE.Destroy;
        E.NextE.Destroy;
        E.Destroy;
      end;
      FFirstT := T.NextT;
      T.Destroy;
    end;
  for i := 0 to Points.Count - 1 do
    with TDPoint(Points[i]) do
    begin
      Clear;
      if DelP then Destroy;
    end;
  Points.Clear;
  FHullStart := nil;
end;

procedure TDelaunay.DeleteTri(T: TDTriangle);
begin
  with T do
  begin
    if PriorT = nil then FFirstT := NextT else PriorT.FNextT := NextT;
    if NextT <> nil then NextT.FPriorT := PriorT;
    T.Destroy;
  end;
end;

procedure TDelaunay.AddTri(T: TDTriangle);
begin
  with T do
  begin
    FNextT := FirstT;
    if NextT <> nil then
    begin
      NextT.FPriorT := T;
      Index := NextT.Index - 1;
    end else
      Index := 0;
    FFirstT := T;
  end;
  if T.Index = 1 then
  begin
    while T.NextT <> nil do T := T.NextT;
    T.Index := 0;
    while T.PriorT <> nil do
    begin
      T := T.PriorT;
      T.Index := T.NextT.Index - 1;
    end;
  end;
end;

procedure TDelaunay.UpdateTri(T: TDTriangle; P: TDPoint);
var
  e1, e2, e3, e4, e5, e6, e1new, e2new, e3new, e4new: TDEdge;
  b1, b2, b3: Boolean;
begin
  e1 := T.E; e2 := e1.NextE; e3 := e2.NextE;
  b1 := e1.OnSide(P) = 0;
  b2 := e2.OnSide(P) = 0;
  b3 := e3.OnSide(P) = 0;
  if not(b1 or b2 or b3) then
  begin
    e1new := TDEdge.Create(e1.LeftP, P);
    e2new := TDEdge.Create(e2.LeftP, P);
    e3new := TDEdge.Create(e3.LeftP, P);
    DeleteTri(T);     // remove old triangle
    AddTri(TDTriangle.Create(e1, e2new, e1new.CreateInv));
    AddTri(TDTriangle.Create(e2, e3new, e2new.CreateInv));
    AddTri(TDTriangle.Create(e3, e1new, e3new.CreateInv));
    SwapTri(e1);   // swap test for the three new triangles
    SwapTri(e2);
    SwapTri(e3);
    exit;
  end;
  if b2 then e1 := e2 else if b3 then e1 := e3;
  e2 := e1.NextE; e3 := e2.NextE;
  e4 := e1.InvE;
  if e4 = nil then     // one triangle involved
  begin
    e3new := TDEdge.Create(e3.LeftP, P);
    e4 := TDEdge.Create(P, e2.LeftP);
    e1new := TDEdge.Create(e1.LeftP, P);
    e5 := e3new.CreateInv;
    e1new.SetPointRefer;
    DeleteTri(T);                   // remove oldtriangle
    // add two new triangles
    AddTri(TDTriangle.Create(e2, e3new, e4));
    AddTri(TDTriangle.Create(e3, e1new, e5));
    e1new.LeftP.LastE.FNextH := e1new;
    e1new.FNextH := e4;
    e4.FNextH := e1.NextH;
    FHullStart := e4;
    e1.Destroy;
    SwapTri(e2);   // swap test for the two new triangles
    SwapTri(e3);
    SwapTri(e3new);
  end else begin        // two triangle involved
    e5 := e4.NextE; e6 := e5.NextE;
    e1new := TDEdge.Create(e1.LeftP, p);
    e2new := TDEdge.Create(e2.LeftP, p);
    e3new := TDEdge.Create(e3.LeftP, p);
    e4new := TDEdge.Create(e6.LeftP, p);
    DeleteTri(T);                   // remove oldtriangle
    DeleteTri(e4.T);               // remove old triangle
    e5.SetPointRefer;   // because e, e4 removed, reset edge index of node p1 and p2
    e2.SetPointRefer;
    AddTri(TDTriangle.Create(e2, e3new, e2new.CreateInv));
    AddTri(TDTriangle.Create(e3, e1new, e3new.CreateInv));
    AddTri(TDTriangle.Create(e5, e4new, e1new.CreateInv));
    AddTri(TDTriangle.Create(e6, e2new, e4new.CreateInv));
    e1.Destroy;
    e4.Destroy;
    SwapTri(e2);   // swap test for the three new triangles
    SwapTri(e3);
    SwapTri(e5);
    SwapTri(e6);
    SwapTri(e1new);
    SwapTri(e2new);
    SwapTri(e3new);
    SwapTri(e4new);
  end;
end;

procedure TDelaunay.UpdateHull(P: TDPoint; Ind: Integer);
var
  EH, E1, E2, E3, E, FirstE, LastE: TDEdge;
  Left: Boolean;
  i: Integer;
begin
  E := HullStart;
  with Points do
    if FirstT = nil then
    begin
      if E = nil then
      begin  // First edge
        FHullStart := TDEdge.Create(List[0], TDPoint(List[0]).NextOrig);
        HullStart.RightP.FE := HullStart;
        Exit;
      end;

      i := E.OnSide(P);
      if i = 0 then
      begin         // Update line
        if Ind = 0 then E1 := nil else E1 := TDPoint(List[Ind - 1]).E;
        if Ind = Count - 1 then E2 := nil else E2 := TDPoint(List[Ind + 1]).FE;
        if E2 <> nil then
        begin
          if E1 <> nil then
          begin
            E := TDEdge.Create(E1.LeftP, P);
            E1.Update(P, E1.RightP);
            E.FNextE := E1.NextE;
            E1.FNextE := E;
          end else begin
            E := TDEdge.Create(P, E2.LeftP);
            E2.FNextE := E;
          end;
        end else begin
          E := TDEdge.Create(E1.RightP, P);
          P.FE := E;
          E.FNextE := E1;
          FHullStart := E;
        end;
        Exit;
      end;

      with E do  // Create first triangles
      begin
        EH := NextE;
        if i = -1 then Update(RightP, LeftP);
        E2 := TDEdge.Create(RightP, P);
        E3 := TDEdge.Create(P, LeftP);
        AddTri(TDTriangle.Create(E, E2, E3));
        FNextH := E2; E2.FNextH := E3; E3.FNextH := E;
      end;
      while EH <> nil do
      begin
        E := EH.NextE;
        UpdateHull(EH.LeftP, 0);
        EH.Destroy;
        EH := E;
      end;
      Exit;
    end;

  FirstE := nil; LastE := nil;
  Left := False;
  repeat
    if E.OnSide(P) = -1 then
    begin
      if not Left and (LastE <> nil) then Break;
      LastE := E;
      Left := True;
    end else begin
      if Left and (FirstE <> nil) then Break;
      FirstE := E;
      Left := False;
    end;
    E := E.NextH;
  until E = HullStart;

  EH := FirstE.NextH;
  E3 := nil;
  repeat
    E := EH;
    EH := EH.NextH;
    E2 := TDEdge.Create(E.LeftP, P);
    if E3 <> nil then E3.SetInvE(E2) else FirstE.FNextH := E2;
    E3 := TDEdge.Create(P, E.RightP);
    AddTri(TDTriangle.Create(E.CreateInv, E2, E3));
    SwapTri(E);
  until E = LastE;
  FHullStart := FirstE;
  FirstE.NextH.FNextH := E3;
  E3.FNextH := LastE.NextH;
end;

function TDelaunay.FindTri(P: TDPoint): TDTriangle;
var
  e: TDEdge;
begin
  Result := nil;
  if FirstT = nil then Exit;
  e := FirstT.E;
  if e.OnSide(P) = -1 then e := e.InvE;
  while (e <> nil) do
  begin
    e := e.NextE;
    if e.OnSide(P) <> -1 then
    begin
      e := e.NextE;
      if e.OnSide(P) <> -1 then
      begin
        Result := e.T;
        Exit;
      end;
    end;
    e := e.InvE;
  end;
end;

function TDelaunay.FindTriangle(x, y: Integer): TDTriangle;
var
  p: TDPoint;
begin
  p := TDPoint.Create(x, y);
  try
    Result := FindTri(p);
  finally
    p.Destroy;
  end;
end;

procedure TDelaunay.SwapTri(e11: TDEdge);
var
  e21, e12, e13, e22, e23: TDEdge;
begin
  e21 := e11.InvE;
  if (e21 = nil) or (e21.T = nil) then exit;
  e12 := e11.NextE; e13 := e12.NextE; e22 := e21.NextE; e23 := e22.NextE;
  if (e11.T.InCircle(e22.RightP)) or (e21.T.InCircle(e12.RightP)) then
  begin
    e11.Update(e22.RightP, e12.RightP);
    e21.Update(e12.RightP, e22.RightP);
    e11.SetInvE(e21);
    e13.T.Update(e13, e22, e11);
    e23.T.Update(e23, e12, e21);
    e12.SetPointRefer;
    e22.SetPointRefer;
    SwapTri(e12);
    SwapTri(e22);
    SwapTri(e13);
    SwapTri(e23);
  end;
end;

function TDelaunay.Insert(px, py: Integer): TDPoint;
begin
  Result := TDPoint.Create(px, py);
  InsertPoint(Result);
end;

function TDelaunay.RightOrder(X, Y: Integer): Integer;
var
  L, R, S: Integer;
  P: TDPoint;
begin
  with Points do
  begin
    L := 0;
    R := Count - 1;
    while R >= L do
    begin
      S := (R + L) div 2;
      P := List[S];
      if (P.X < X) then L := S + 1 else
        if (P.X > X) then R := S - 1 else
          if (P.Y < Y) then L := S + 1 else
            if (P.Y > Y) then R := S - 1 else
            begin
              P := P.Original;
              while List[S] <> P do dec(S);
              Result := S;
              exit;
            end;
    end;
    Result := R;
  end;
end;

function TDelaunay.ByCoord(X, Y: Integer): TDPoint;
var
  i: Integer;
begin
  i := RightOrder(X, Y);
  if (i <> -1) then
  begin
    Result := Points.List[i];
    if (X = Result.X) and (Y = Result.Y) then Exit;
  end;
  Result := nil;
end;

function TDelaunay.Nearest(X, Y: Integer): TDPoint;
var
  Min, S: Double;
//  T: TDTriangle;
//  E: TDEdge;
  i: Integer;
  P: TDPoint;
begin
  with Points do
  begin
    Result := nil;
    if Count = 0 then Exit;
    Result := List[0];
    Min := Result.Distance(X, Y);
    for i := 1 to Count - 1 do
    begin
      P := List[i];
      S := P.Distance(X, Y);
      if S < Min then
      begin
        Result := P;
        Min := S;
      end;
    end;
  end;

{  T := FindTriangle(X, Y);
  if T = nil then
  begin
    Result := nil;
    if Points.Count <> 0 then
      if FHullStart <> nil then
      begin
        E := FHullStart;
        Result := E.LeftP;
        Min := Result.Distance(X, Y);
        repeat
          S := E.RightP.Distance(X, Y);
          if S < Min then
          begin
            Result := E.RightP;
            Min := S;
          end;
          E := E.NextH;
        until (E = nil) or (E = FHullStart);
      end else
        Result := Points[0];
    exit;
  end;

  E := T.E;
  with T.E do
  begin
    Min := 0;
    Result := nil;
    for i := 1 to 3 do
    begin
      S := E.RightP.Distance(X, Y);
      if (S < Min) or (Result = nil) then
      begin
        Result := E.RightP;
        Min := S;
      end;
      if E.InvE <> nil then
        with E.InvE.NextE do
        begin
          S := RightP.Distance(X, Y);
          if S < Min then
          begin
            Result := RightP;
            Min := S;
          end;
        end;
      E := E.NextE;
    end;
  end;}
end;

{function SCompare(p1, p2: Pointer): Integer;
begin
  Result := -1;
  if p1 = p2 then Result := 0 else
    if p2 <> nil then
      with TDPoint(p1) do
        if (p1 = nil) or (X > TDPoint(p2).X) or
           ((X = TDPoint(p2).X) and ((Y > TDPoint(p2).Y) or
           ((Y = TDPoint(p2).Y) and (FIndex > TDPoint(p2).FIndex)))) then
           Result := 1;
end;
}

function TDelaunay.GetPoint(Index: Integer): TDPoint;
begin
  with Points do
  begin
    if (Index < 0) or (Index >= Count) then
      raise EListError.CreateFmt('?', [Index]);
    Result := List[Index];
  end;
end;

function TDelaunay.GetPointsCount: Integer;
begin
  Result := Points.Count;
end;

function TDelaunay.Exists(X, Y: Integer): Boolean;
begin
  Result := ByCoord(X, Y) <> nil;
end;

procedure TDelaunay.DeleteDupl;
var
  i: Integer;
  P: TDPoint;
begin
  with Points do
  begin
    for i := Count - 1 downto 0 do
    begin
      P := List[i];
      if not P.IsOriginal then P.Destroy;
    end;
  end;
end;

procedure TDelaunay.InsertPoint(P: TDPoint);
var
  i: Integer;
  T: TDTriangle;
begin
  with P, Points do
  begin
    if Owner <> nil then exit;
    FOwner := Self;
    i := RightOrder(X, Y);
    Insert(i + 1, P);
    if (i <> -1) and IsIdent(List[i]) then FOriginal := List[i];
    if IsOriginal and (Count <> 1) then
    begin
      T := FindTri(P);
      if T = nil then UpdateHull(P, i + 1) else UpdateTri(T, P);
    end;
  end;
  if Assigned(FOnInsert) then OnInsert(Self, P);
  BreakEnum;
end;

procedure TDelaunay.RemovePoint(P: TDPoint);
var
  i, Ind: Integer;
  P1, P2: TDPoint;
  SE, LE: TDEdge;
begin
  with P, Points do
  begin
    if Owner <> Self then exit;
    Ind := P.Index;
    if IsOriginal then
    begin
      if (Ind = Count - 1) or TDPoint(List[Ind + 1]).IsOriginal then
        Self.Delete(P)
      else begin
        P1 := List[Ind + 1];
        SE := FirstE; LE := LastE;
        if SE <> nil then
          repeat
            SE.ChangeP(P, P1);
            if SE.InvE <> nil then SE.InvE.ChangeP(P, P1);
            if SE = LE then Break;
            SE := SE.NextPE;
          until False;
        P1.FE := FE;
        P1.FOriginal := P1;
        for i := Ind + 2 to Count - 1 do
        begin
          P2 := List[i];
          if P2.IsOriginal then Break else P2.FOriginal := P1;
        end;
      end;
    end;
    Delete(Ind);
    P.Clear;
  end;
  if Assigned(FOnRemove) then OnRemove(Self, P);
  FBreakEnum := True;
end;

procedure TDelaunay.Delete(P: TDPoint);
var
  E, StartE, E1, LastE: TDEdge;
begin
  if FirstT = nil then
  begin  // Triangles doesn't exist, update line
    E1 := P.FE;
    if E1 = nil then Exit;
    E := E1.NextE;
    if (P <> E1.RightP) and (E <> nil) then
    begin
      E1.Update(E.LeftP, E1.RightP);
      E1.FNextE := E.NextE;
      E.Destroy;
      Exit;
    end;
    E1.RightP.FE.FNextE := nil;
    if HullStart = E1 then
    begin
      E1.LeftP.FE := E;
      E1.RightP.FE := nil;
      FHullStart := E;
    end;
    E1.Destroy;
    Exit;
  end;

  E := P.FirstE;
  StartE := E.NextE;
  if E.InvE = nil then E.Destroy;
  E := StartE;
  repeat
    DeleteTri(E.T);
    E.FT := nil;
    E1 := E.NextE.InvE;
    E.NextE.Destroy;
    E.FNextE := nil;
    E.SetPointRefer;
    if E1 = nil then
    begin
      DeleteOnHull(StartE);
      Exit;
    end;
    E.FNextE := E1.NextE;
    E := E1.NextE;
    E1.Destroy;
  until E = StartE;

  // Point is inner
  LastE := E; E := E.NextE;
  while E.NextE.NextE <> LastE do
  begin
    with E.NextE do
      if (E.OnSide(RightP) = 1) then
      begin
        E1 := TDEdge.Create(RightP, E.LeftP);
        StartE := NextE;
        while StartE <> LastE do
          if (E.OnSide(StartE.RightP) <> -1) and
             (OnSide(StartE.RightP) <> -1) and
             (E1.OnSide(StartE.RightP) <> -1) then
            Break else StartE := StartE.NextE;
        if StartE <> LastE then
        begin
          E1.Destroy;
          NextE.SetPointRefer;
        end else begin
          E1 := E1.CreateInv;
          E1.FNextE := NextE;
          AddTri(TDTriangle.Create(E, E.NextE, E1.InvE));
          LastE.FNextE := E1;
          SwapTri(E);
          SwapTri(E.NextE);
          E := E1;
        end;
      end;
    LastE := E;
    E := E.NextE;
{    if E.NextE = LastE then
    begin
      E.InvE.SetInvE(LastE.InvE);
      E.InvE.SetPointRefer;
      LastE.InvE.SetPointRefer;
      SwapTri(E.InvE);
      Exit;
    end;}
  end;
  AddTri(TDTriangle.Create(E, E.NextE, LastE));
  SwapTri(E);
  SwapTri(E.NextE);
  SwapTri(LastE);
end;

procedure TDelaunay.DeleteOnHull(StartE: TDEdge);
var
  E, E1, LastE: TDEdge;
begin
  E := StartE;
  LastE := nil;
  while E.NextE <> nil do
  begin
    if E.OnSide(E.NextE.RightP) = 1 then
    begin
      E1 := TDEdge.Create(E.NextE.RightP, E.LeftP).CreateInv;
      E1.FNextE := E.NextE.NextE;
      AddTri(TDTriangle.Create(E, E.NextE, E1.InvE));
      SwapTri(E);
      E := E1;
      if LastE <> nil then
      begin
        E := LastE;
        LastE := LastE.NextE;
        E.FNextE := E1;
      end;
    end else begin
      E1 := E.NextE;
      E.FNextE := LastE;
      LastE := E;
      E := E1;
    end;
  end;
  E.FNextE := LastE;

  // Reconstruct the convex hull
  if FirstT = nil then
  begin                 // Reconstruct line
    if E.RightP.Index < StartE.LeftP.Index then
    begin
       LastE := nil;
       repeat
         E1 := E.NextE;
         E.FNextE := LastE;
         E.Update(E.RightP, E.LeftP);
         LastE := E;
         if E1 = nil then Break;
         E := E1;
       until False;
    end;
    FHullStart := E;
    E.RightP.FE := E;
    Exit;
  end;

  FHullStart := E.InvE;
  repeat
    E1 := E.NextE;
    with E.InvE do
    begin
      SetInvE(nil);
      NextE.SetPointRefer;
      if E1 = nil then FNextH := RightP.FirstE else FNextH := E1.InvE;
    end;
    E.Destroy;
    E := E1;
  until E = nil;
  HullStart.SetPointRefer;
  HullStart.LeftP.LastE.FNextH := HullStart;
end;

procedure TDelaunay.BreakEnum;
begin
  FBreakEnum := true;
end;

procedure TDelaunay.EnumPoints(Dupl: Boolean);
var
  i: Integer;
  P: TDPoint;
begin
  if not Assigned(FOnEnumPoint) then Exit;
  FBreakEnum := False;
  with Points do
    if Dupl then
      for i := 0 to Count - 1 do
      begin
        P := List[i];
        OnEnumPoint(Self, P);
        if FBreakEnum then Break;
      end
    else
      for i := 0 to Count - 1 do
      begin
        P := List[i];
        if P.IsOriginal then OnEnumPoint(Self, P);
        if FBreakEnum then Break;
      end
end;

procedure TDelaunay.EnumEdges(BothSide: Boolean);
var
  i: Integer;
  E: TDEdge;
  T: TDTriangle;
begin
  if not Assigned(FOnEnumEdge) then Exit;
  FBreakEnum := False;
  if FirstT = nil then
  begin
    E := HullStart;
    while E <> nil do
    begin
      OnEnumEdge(Self, E);
      if FBreakEnum then Exit;
      E := E.NextE;
    end;
    Exit;
  end;
  T := FirstT;
  repeat
    E := T.E;
    for i := 1 to 3 do
    begin
      if BothSide or E.IsPrior then
      begin
        OnEnumEdge(Self, E);
        if FBreakEnum then Exit;
      end;
      E := E.NextE;
    end;
    T := T.NextT;
  until T = nil;
end;

procedure TDelaunay.EnumTriangles;
var
  T: TDTriangle;
begin
  if not Assigned(FOnEnumTrinagle) then Exit;
  FBreakEnum := False;
  T := FirstT;
  while T <> nil do
  begin
    OnEnumTriangle(Self, T);
    if FBreakEnum then Break;
    T := T.NextT;
  end;
end;

procedure TDelaunay.EnumHEdges;
var
  E: TDEdge;
begin
  if not Assigned(FOnEnumHEdge) then Exit;
  FBreakEnum := False;
  E := HullStart;
  while (E <> nil) do
  begin
    OnEnumHEdge(Self, E);
    if FBreakEnum then Break;
    if E.T = nil then E := E.NextE else E := E.NextH;
    if E = HullStart then Break;
  end;
end;

procedure TDelaunay.GetPList(L: TList);
var i: Integer;
begin
  with Points do
  begin
    i := L.Count;
    L.Count := L.Count + Count;
    System.Move(List^[0], L.List^[i], SizeOf(Pointer) * Count);
  end;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TDelaunay]);
end;

end.
